@@ -1,9 +1,28 @@
+// This file is part of Oaklisp.
+//
+// This program is free software; you can redistribute it and/or
+// modify it under the terms of the GNU General Public License as
+// published by the Free Software Foundation; either version 2 of the
+// License, or (at your option) any later version.
+//
+// This program is distributed in the hope that it will be useful,
+// but WITHOUT ANY WARRANTY; without even the implied warranty of
+// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+// GNU General Public License for more details.
+//
+// The GNU GPL is available at http://www.gnu.org/licenses/gpl.html
+// or from the Free Software Foundation, 59 Temple Place - Suite 330,
+// Boston, MA 02111-1307, USA
+
+
 /**********************************************************************
- *     Copyright (c) by Barak Pearlmutter and Kevin Lang, 1987-99.    *
- *     Copyright (c) by Alex Stuebinger, 1998-99.                     *
- *     Distributed under the GNU General Public License v2 or later   *
+ *  Copyright (c) by Barak A. Pearlmutter and Kevin J. Lang, 1987-99. *
+ *  Copyright (c) by Alex Stuebinger, 1998-99.                        *
+ *  Distributed under the GNU General Public License v2 or later      *
  **********************************************************************/
 
+#define _REENTRANT
+
 #include <stddef.h>
 #include <stdlib.h>
 #include <stdio.h>
@@ -14,12 +33,14 @@
 #include "gc.h"
 
 
-
 #ifdef USE_VADVISE
 #include <sys/vadvise.h>
 #endif
 
 
+#define FORTHREADS THREADY( for (my_index=0; my_index<next_index; my_index++) )
+
+
 
 /* 1/RECLAIM_FACTOR is the target for how much of new space should 
    be used after a gc.  If more than this is used, the next new 
@@ -38,12 +59,13 @@ unsigned long loc_transport_count;
 #define GC_EXAMINE_BUFFER_SIZE 16
 #endif
 
+#ifdef THREADS
+ref_t gc_examine_buffer_array[MAX_THREAD_COUNT][GC_EXAMINE_BUFFER_SIZE];
+ref_t *gc_examine_ptr_array[MAX_THREAD_COUNT];
+#else
 ref_t gc_examine_buffer[GC_EXAMINE_BUFFER_SIZE];
 ref_t *gc_examine_ptr = gc_examine_buffer;
-
-
-
-
+#endif
 
 #define GC_TOUCH(x)			\
 {					\
@@ -79,7 +101,6 @@ ref_t *gc_examine_ptr = gc_examine_buffer;
   (x) = LOC_TO_PTR(loc_touch0(PTR_TO_LOC(x),1));	\
 }
 
-
 void
 printref(FILE * fd, ref_t refin)
 {
@@ -385,7 +406,17 @@ pc_touch(u_int16_t * o_pc)
 {
   ref_t *pcell = (ref_t *) ((unsigned long)o_pc & ~TAG_MASKL);
 
+  /*
+    It is possible that the gc was called while a vm was executing the last
+    instruction in a code block (hopefully a branch or funcall) in
+    a multithreaded enviornment.  So let's back up the pc one before gc'ing
+    it.  However, this means the gc generally should not be called until the
+    loop has read at least one instruction in the code block.
+  */
+  /* pcell--;  Changed my mind.  Moved POLL_GC_SIGNALS to top of loop. */
   LOC_TOUCH_PTR(pcell);
+  /* pcell++; */
+
   return
     (u_int16_t *) ((u_int32_t) pcell
 		   | ((u_int32_t) o_pc & TAG_MASK));
@@ -407,7 +438,32 @@ gc(bool pre_dump, bool full_gc, char *reason, size_t amount)
  */
 {
   long old_taken;
+  long old_spatic_taken;
   ref_t *p;
+#ifdef THREADS
+  bool ready=false;
+  int my_index;
+  int i;
+  int *my_index_p;
+  my_index_p = pthread_getspecific (index_key);
+  my_index = *my_index_p;
+  gc_ready[my_index] = 1;
+  set_gc_flag (true);
+#endif
+
+#ifdef THREADS
+  /*Problem here is next_index could change if someone creates a thread
+    while someone else is gc'ing*/
+   while (ready == false) {
+    ready = true;
+    for (i = 0; i < next_index; i++) {
+      if (gc_ready[i] == 0) {
+          ready = false;
+          break;
+      }
+    }
+  }
+#endif
 
   /* The full_gc flag is also a global to avoid ugly parameter passing. */
   set_external_full_gc(full_gc);
@@ -420,17 +476,23 @@ gc_top:
 
   if (trace_gc > 2 && !pre_dump)
     {
-      fprintf(stderr, "value ");
-      dump_stack(&value_stack);
-      fprintf(stderr, "context ");
-      dump_stack(&context_stack);
+      FORTHREADS {
+	fprintf (stderr, "value ");
+	dump_stack (value_stack_address);
+	fprintf (stderr, "context ");
+	dump_stack (context_stack_address);
+      }
     }
   if (trace_gc > 1)
     fprintf(stderr, "; Flipping...");
 
   old_taken = free_point - new_space.start;
+  old_spatic_taken = spatic.size;
   old_space = new_space;
 
+  if (trace_gc > 2)
+    fprintf(stderr, "old taken: %ld", old_taken);
+
   if (full_gc)
     new_space.size += spatic.size;
   else
@@ -439,6 +501,7 @@ gc_top:
   alloc_space(&new_space, new_space.size);
   free_point = new_space.start;
 
+
   transport_count = 0;
 
   if (trace_gc > 1)
@@ -458,37 +521,41 @@ gc_top:
 	GC_TOUCH(e_loc_type);
 	GC_TOUCH(e_cons_type);
 	GC_TOUCH_PTR(e_subtype_table, 2);
-	/* e_bp is a locative, but a pointer to the object should
-	   exist, so we need only touch it in the locative pass. */
-	GC_TOUCH_PTR(e_env, 0);
 	/* e_nargs is a fixnum.  Nor is it global... */
-	GC_TOUCH(e_env_type);
-	GC_TOUCH_PTR(e_argless_tag_trap_table, 2);
-	GC_TOUCH_PTR(e_arged_tag_trap_table, 2);
-	GC_TOUCH(e_object_type);
-	GC_TOUCH(e_segment_type);
-	GC_TOUCH(e_code_segment);
-	GC_TOUCH(e_current_method);
-	GC_TOUCH(e_uninitialized);
-	GC_TOUCH(e_method_type);
-	GC_TOUCH(e_operation_type);
-
-	for (p = gc_examine_buffer; p < gc_examine_ptr; p++)
-	  GC_TOUCH(*p);
-
-
+	GC_TOUCH (e_env_type);
+	GC_TOUCH_PTR (e_argless_tag_trap_table, 2);
+	GC_TOUCH_PTR (e_arged_tag_trap_table, 2);
+	GC_TOUCH (e_object_type);
+	GC_TOUCH (e_segment_type);
+	FORTHREADS {
+	  /* e_bp is a locative, but a pointer to the object should exist, so we
+	     need only touch it in the locative pass. */
+	  GC_TOUCH_PTR(e_env, 0);
+	  GC_TOUCH (e_code_segment);
+	  GC_TOUCH (e_current_method);
+	  GC_TOUCH (e_process);
+	}
+	GC_TOUCH (e_uninitialized);
+	GC_TOUCH (e_method_type);
+	GC_TOUCH (e_operation_type);
 
+	FORTHREADS {
+	  for (p = gc_examine_buffer; p < gc_examine_ptr; p++)
+	    GC_TOUCH(*p);
+	}
 
 	/* Scan the stacks. */
-	for (p = value_stack.bp; p <= value_stack.sp; p++)
-	  GC_TOUCH(*p);
+	FORTHREADS {
+	  for (p = value_stack.bp; p <= value_stack.sp; p++)
+	    GC_TOUCH(*p);
 
-	for (p = context_stack.bp; p <= context_stack.sp; p++)
-	  GC_TOUCH(*p);
+	  for (p = context_stack.bp; p <= context_stack.sp; p++)
+	    GC_TOUCH(*p);
 
-	/* Scan the stack segments. */
-	GC_TOUCH(value_stack.segment);
-	GC_TOUCH(context_stack.segment);
+	  /* Scan the stack segments. */
+	  GC_TOUCH(value_stack.segment);
+	  GC_TOUCH(context_stack.segment);
+	}
 
 	/* Scan static space. */
 	if (!full_gc)
@@ -513,19 +580,22 @@ gc_top:
 
     if (!pre_dump)
       {
-	LOC_TOUCH_PTR(e_bp);
-	e_pc = pc_touch(e_pc);
+	FORTHREADS {
+	  LOC_TOUCH_PTR (e_bp);
+	  e_pc = pc_touch (e_pc);
 
-	LOC_TOUCH(e_uninitialized);
+	  LOC_TOUCH(e_uninitialized);
 
-	for (p = gc_examine_buffer; p < gc_examine_ptr; p++)
-	  LOC_TOUCH(*p);
+	  for (p = gc_examine_buffer; p < gc_examine_ptr; p++)
+	    LOC_TOUCH(*p);
 
-	for (p = value_stack.bp; p <= value_stack.sp; p++)
-	  LOC_TOUCH(*p);
+	  for (p = value_stack.bp; p <= value_stack.sp; p++)
+	    LOC_TOUCH(*p);
 
-	for (p = context_stack.bp; p <= context_stack.sp; p++)
-	  LOC_TOUCH(*p);
+	  for (p = context_stack.bp; p <= context_stack.sp; p++)
+
+	    LOC_TOUCH(*p);
+	}
 
 	/* Scan spatic space. */
 	if (!full_gc)
@@ -565,41 +635,46 @@ gc_top:
 
     if (!pre_dump)
       {
-	GGC_CHECK(e_t);
-	GGC_CHECK(e_fixnum_type);
-	GGC_CHECK(e_loc_type);
-	GGC_CHECK(e_cons_type);
-	GC_CHECK(PTR_TO_REF(e_subtype_table - 2), "e_subtype_table");
-	GC_CHECK(PTR_TO_LOC(e_bp), "PTR_TO_LOC(e_bp)");
-	GC_CHECK(PTR_TO_REF(e_env), "e_env");
+	GGC_CHECK (e_t);
+	GGC_CHECK (e_fixnum_type);
+	GGC_CHECK (e_loc_type);
+	GGC_CHECK (e_cons_type);
+	GC_CHECK (PTR_TO_REF (e_subtype_table - 2), "e_subtype_table");
+	FORTHREADS {
+	  GC_CHECK (PTR_TO_LOC (e_bp), "PTR_TO_LOC(E_BP)");
+	  GC_CHECK (PTR_TO_REF (e_env), "e_env");
+        }
 	/* e_nargs is a fixnum.  Nor is it global... */
-	GGC_CHECK(e_env_type);
-	GC_CHECK(PTR_TO_REF(e_argless_tag_trap_table - 2),
-		 "e_argless_tag_trap_table");
-	GC_CHECK(PTR_TO_REF(e_arged_tag_trap_table - 2),
-		 "e_arged_tag_trap_table");
-	GGC_CHECK(e_object_type);
-	GGC_CHECK(e_segment_type);
-	GGC_CHECK(e_code_segment);
-	GGC_CHECK(e_current_method);
-	GGC_CHECK(e_uninitialized);
-	GGC_CHECK(e_method_type);
-	GGC_CHECK(e_operation_type);
+	GGC_CHECK (e_env_type);
+	GC_CHECK (PTR_TO_REF (e_argless_tag_trap_table - 2), "e_argless_tag_trap_table");
+	GC_CHECK (PTR_TO_REF (e_arged_tag_trap_table - 2), "e_arged_tag_trap_table");
+	GGC_CHECK (e_object_type);
+	GGC_CHECK (e_segment_type);
+	FORTHREADS {
+	  GGC_CHECK (e_code_segment);
+	  GGC_CHECK (e_current_method);
+	  GGC_CHECK (e_process);
+	}
+	GGC_CHECK (e_uninitialized);
+	GGC_CHECK (e_method_type);
+	GGC_CHECK (e_operation_type);
 
 	/* Scan the stacks. */
-	for (p = value_stack.bp; p <= value_stack.sp; p++)
-	  GC_CHECK1(*p, "value_stack.bp[%d] = ",
-		    (long)(p - value_stack.bp));
+	FORTHREADS {
+	  for (p = value_stack.bp; p <= value_stack.sp; p++)
+	    GC_CHECK1(*p, "value_stack.bp[%d] = ",
+		      (long)(p - value_stack.bp));
 
-	for (p = context_stack.bp; p <= context_stack.sp; p++)
-	  GC_CHECK1(*p, "context_stack.bp[%d] = ",
-		    (long)(p - context_stack.bp));
+	  for (p = context_stack.bp; p <= context_stack.sp; p++)
+	    GC_CHECK1(*p, "context_stack.bp[%d] = ",
+		      (long)(p - context_stack.bp));
 
-	GGC_CHECK(value_stack.segment);
-	GGC_CHECK(context_stack.segment);
+	  GGC_CHECK(value_stack.segment);
+	  GGC_CHECK(context_stack.segment);
 
-	/* Make sure the program counter is okay. */
-	GC_CHECK((ref_t) ((ref_t) e_pc | LOC_TAG), "e_pc");
+	  /* Make sure the program counter is okay. */
+	  GC_CHECK ((ref_t) ((ref_t) e_pc | LOC_TAG), "e_pc");
+	}
       }
     /* Scan the heap. */
 
@@ -613,7 +688,12 @@ gc_top:
 #endif /* not defined(FAST) */
 
   /* Hopefully there are no more references into old space. */
-  free_space(&old_space);
+  if (!pre_dump)
+    free_space(&old_space);
+
+  if (!pre_dump && full_gc)
+    free_space(&spatic);
+
 
 
 #ifdef USE_VADVISE
@@ -625,14 +705,19 @@ gc_top:
 
   if (trace_gc > 2 && !pre_dump)
     {
-      fprintf(stderr, "value_stack ");
-      dump_stack(&value_stack);
-      fprintf(stderr, "context_stack ");
-      dump_stack(&context_stack);
+      FORTHREADS {
+#ifdef THREADS
+	fprintf (stderr, "Thread %d\n", my_index);
+#endif
+	fprintf (stderr, "value_stack ");
+	dump_stack (value_stack_address);
+	fprintf (stderr, "context_stack ");
+	dump_stack (context_stack_address);
+      }
     }
   {
     long new_taken = free_point - new_space.start;
-    long old_total = old_taken + (full_gc ? spatic.size : 0);
+    long old_total = old_taken + (full_gc ? old_spatic_taken : 0);
     long reclaimed = old_total - new_taken;
 
     if (trace_gc == 1)
@@ -643,7 +728,7 @@ gc_top:
       {
 	fprintf(stderr, "; GC complete.  %ld ", old_total);
 	if (full_gc)
-	  fprintf(stderr, "(%ld+%ld) ", (long)spatic.size, (long)old_taken);
+	  fprintf(stderr, "(%ld+%ld) ", old_spatic_taken, old_taken);
 	fprintf(stderr, "compacted to %ld; %ld (%ld%%) garbage.\n",
 		new_taken, reclaimed, (100 * reclaimed) / old_total);
       }
@@ -664,7 +749,6 @@ gc_top:
 	  case 1:
 	    fprintf(stderr, ",resize:%ld", (long)e_next_newspace_size);
 	    break;
-
 	  default:
 	    fprintf(stderr,
 		    "; Expanding next new space from %ld to %ld (%ld%%).\n",
@@ -690,15 +774,10 @@ gc_top:
 	  }
       }
 
-    if (full_gc)
-      free_space(&spatic);
-
     if (full_gc && !pre_dump)
       {
-	/* Ditch old spatic, move _new to spatic, and reallocate new. */
-	/* This is a bug
-	   free_space (&spatic);
-	 */
+	/* move _new to spatic, and reallocate new. */
+
 	spatic = new_space;
 	realloc_space(&spatic, free_point - new_space.start);
 
@@ -737,6 +816,12 @@ gc_top:
     if (trace_gc)
       fflush(stdout);
   }
+#ifdef THREADS
+    my_index_p = pthread_getspecific (index_key);
+    my_index = *my_index_p;
+    gc_ready[my_index] = 0;
+    set_gc_flag (false);
+#endif 
 }
 
 
