@@ -1,7 +1,24 @@
+// This file is part of Oaklisp.
+//
+// This program is free software; you can redistribute it and/or
+// modify it under the terms of the GNU General Public License as
+// published by the Free Software Foundation; either version 2 of the
+// License, or (at your option) any later version.
+//
+// This program is distributed in the hope that it will be useful,
+// but WITHOUT ANY WARRANTY; without even the implied warranty of
+// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+// GNU General Public License for more details.
+//
+// The GNU GPL is available at http://www.gnu.org/licenses/gpl.html
+// or from the Free Software Foundation, 59 Temple Place - Suite 330,
+// Boston, MA 02111-1307, USA
+
+
 /**********************************************************************
- *     Copyright (c) by Barak Pearlmutter and Kevin Lang, 1987-99.    *
- *     Copyright (c) by Alex Stuebinger, 1998-99.                     *
- *     Distributed under the GNU General Public License v2 or later   *
+ *  Copyright (c) by Barak A. Pearlmutter and Kevin J. Lang, 1987-99. *
+ *  Copyright (c) by Alex Stuebinger, 1998-99.                        *
+ *  Distributed under the GNU General Public License v2 or later      *
  **********************************************************************/
 
 
@@ -13,6 +30,8 @@
    BUG: This can delay interrupt handling when waiting for input.
  */
 
+#define _REENTRANT
+
 #include <stddef.h>
 #include <stdlib.h>
 #include <stdio.h>
@@ -21,15 +40,12 @@
 #include "signal.h"
 
 
-#ifndef _ICC
-
 
 int signal_poll_flag = 0;
 
 static void
 intr_proc(int sig)
 {
-  sig = sig;			/* Keep the compiler from complaining */
   signal_poll_flag++;
 }
 
@@ -58,5 +74,3 @@ clear_signal(void)
   signal_poll_flag = 0;
 }
 #endif /* commented out */
-
-#endif
